# Generated by Django 3.2.5 on 2022-03-31 22:34

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('chit_chat', '0002_room_members_temp'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='room',
            name='members',
        ),
    ]



import json

from rest_framework import exceptions
from channels.generic.websocket import AsyncWebsocketConsumer
from channels.db import database_sync_to_async

from chit_chat.consumer_serializers import ContentSerializer
from .ckc_conf import chat_settings


def async_validation_exception_handler(func):
    async def inner(*args, **kwargs):
        self = args[0]
        assert AsyncWebsocketConsumer in self.__class__.__mro__
        try:
            return await func(*args, **kwargs)
        except exceptions.ValidationError as e:
            errors = e.detail
            if 'non_field_errors' not in errors:
                errors = {'field_errors': errors}
            await self.send(text_data=json.dumps(errors))
        except Exception as e:  # pragma: no cover
            await self.send(text_data=json.dumps({'non_field_errors': ['System Error']}))
            raise e
    return inner


class ChatRoomConsumer(AsyncWebsocketConsumer):
    MESSAGE_TYPES = [
        'chat',
    ]

    async def connect(self):
        user = self.scope['user']
        if not user.is_authenticated:
            await self.close()
            return
        for pk in await self.get_chat_room_pks(user):
            await self.channel_layer.group_add(str(pk), self.channel_name)
        await self.accept()

    async def disconnect(self, close_code):
        user = self.scope['user']
        if user.is_authenticated:
            for pk in await self.get_chat_room_pks(self.scope['user']):
                await self.channel_layer.group_discard(str(pk), self.channel_name)

    async def receive(self, text_data=None, bytes_data=None):
        user = self.scope['user']
        if bytes_data is not None:
            text_data = bytes_data.decode()
        try:
            data = json.loads(text_data)
        except json.decoder.JSONDecodeError:
            await self.send(text_data=json.dumps({'non_field_errors': ['Invalid JSON.']}))
            return

        valid = await self.validate_content(data)
        if valid:
            message_type = data.get('message_type')

            if message_type == 'chat':
                data['user'] = user.pk

                if message := await self.validate_chat_message(data):
                    await self.channel_layer.group_send(
                        str(message.room.pk),
                        {
                            'type': 'chat',
                            'user': user.pk,
                            'room': message.room.pk,
                            'text': message.text,
                            'time': message.created_when.isoformat(),
                        }
                    )

    @async_validation_exception_handler
    async def validate_content(self, data):
        serializer = ContentSerializer(ChatRoomConsumer.MESSAGE_TYPES, data=data)
        return serializer.is_valid(raise_exception=True)

    @async_validation_exception_handler
    @database_sync_to_async
    def validate_chat_message(self, data):
        serializer = chat_settings['SERIALIZERS'].MESSAGE(data=data)
        serializer.is_valid(raise_exception=True)
        return serializer.save()

    async def chat(self, event):
        await self.send(
            text_data=json.dumps({
                'type': 'chat',
                'user': event.get('user'),
                'text': event.get('text'),
                'room': event.get('room'),
                'time': event.get('time'),
            })
        )

    @database_sync_to_async
    def get_chat_room_pks(self, user):
        return list(user.chat_rooms.values_list('pk', flat=True))



from django.contrib.auth import get_user_model
from django.utils import timezone
from django.db import models


User = get_user_model()


class Room(models.Model):
    members = models.ManyToManyField(User, related_name='chat_rooms', through='RoomMembership')
    created_when = models.DateTimeField(default=timezone.now)


class RoomMembership(models.Model):
    room = models.ForeignKey(Room, related_name='memberships', on_delete=models.CASCADE)
    user = models.ForeignKey(User, related_name='chat_room_memberships', on_delete=models.CASCADE)
    archived = models.BooleanField(default=False)
    ignore_notifications = models.BooleanField(default=False)
    created_when = models.DateTimeField(default=timezone.now)


class Message(models.Model):
    text = models.TextField()
    room = models.ForeignKey(Room, related_name='messages', on_delete=models.CASCADE)
    user = models.ForeignKey(User, related_name='chat_room_messages', on_delete=models.CASCADE)
    created_when = models.DateTimeField(default=timezone.now)
    users_who_viewed = models.ManyToManyField(User, related_name='chat_room_messages_viewed')



# Generated by Django 3.2.5 on 2022-03-31 22:29

from django.conf import settings
from django.db import migrations, models


def copy_members_to_temp_row(apps, schema_editor):
    Room = apps.get_model('chit_chat', 'Room')
    RoomMembership = apps.get_model('chit_chat', 'RoomMembership')

    # Ugly code that makes lots of queries .. ok to migrate a few things!
    for room in Room.objects.all():
        for member in room.members.all():
            RoomMembership.objects.create(
                room=room,
                user=member,
            )


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('chit_chat', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='room',
            name='members_temp',
            field=models.ManyToManyField(related_name='chat_rooms_temp', through='chit_chat.RoomMembership', to=settings.AUTH_USER_MODEL),
        ),
        migrations.RunPython(copy_members_to_temp_row)
    ]



import os


DEBUG = True

BASE_DIR = os.path.dirname(__file__)

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        # "NAME": os.path.join(BASE_DIR, "db.sqlite3"),
    }
}

EMAIL_BACKEND = "django.core.mail.backends.console.EmailBackend"

SECRET_KEY = "_"

MIDDLEWARE = (
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)


INSTALLED_APPS = (
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.staticfiles",

    "chit_chat",

    "testapp",
    'channels',
)

AUTH_USER_MODEL = 'testapp.User'

STATIC_URL = "/static/"

ROOT_URLCONF = "urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True
    }
]

# =============================================================================
# DRF
# =============================================================================
REST_FRAMEWORK = {
    # 'DEFAULT_AUTHENTICATION_CLASSES': (
    #     'rest_framework.authentication.TokenAuthentication',
    #     'rest_framework.authentication.SessionAuthentication',
    #     'oauth2_provider.contrib.rest_framework.OAuth2Authentication',
    # ),
    # 'DEFAULT_PERMISSION_CLASSES': (
    #     'rest_framework.permissions.IsAuthenticated',
    # ),
    'DEFAULT_PAGINATION_CLASS': 'rest_framework.pagination.LimitOffsetPagination',
    'PAGE_SIZE': 100,
    # 'DATETIME_INPUT_FORMATS': (
    #     'iso-8601',
    #     '%B %d, %Y',
    # )
}


# =============================================================================
# Channels
# =============================================================================
ASGI_APPLICATION = "routing.application"

CHANNEL_LAYERS = {
    "default": {
        "BACKEND": "channels.layers.InMemoryChannelLayer"
    },
}



from rest_framework import serializers, exceptions
from django.contrib.auth import get_user_model

from chit_chat.models import Room, Message


User = get_user_model()


class ContentSerializer(serializers.Serializer):
    message_type = serializers.CharField()

    def __init__(self, valid_message_types, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.valid_message_types = valid_message_types

    def validate_message_type(self, message_type):
        if message_type not in self.valid_message_types:
            raise exceptions.ValidationError(f'Must be one of the following message types: {self.valid_message_types}')
        return message_type


class ChatMessageSerializer(serializers.Serializer):
    room = serializers.PrimaryKeyRelatedField(queryset=Room.objects.all())
    text = serializers.CharField()
    user = serializers.PrimaryKeyRelatedField(queryset=User.objects.all())

    def validate(self, attrs):
        # Make sure the user is in the room.
        if not attrs['room'].members.filter(pk=attrs['user'].pk).exists():
            raise exceptions.ValidationError('Cannot send messages to chat rooms that you are not a member of.')
        return attrs

    def create(self, validated_data):
        return Message.objects.create(**validated_data)



from django.db import models
from django.contrib.auth.models import AbstractBaseUser, PermissionsMixin


class User(AbstractBaseUser, PermissionsMixin):
    USERNAME_FIELD = 'email'
    REQUIRED_FIELDS = []

    email = models.EmailField(verbose_name="email", max_length=60, unique=True)
    first_name = models.CharField(max_length=64, null=True, blank=True)
    last_name = models.CharField(max_length=64, null=True, blank=True)
    avatar = models.ImageField(upload_to="", null=True)



from rest_framework import serializers

from chit_chat.consumer_serializers import ChatMessageSerializer


class ChatTestSerializer(ChatMessageSerializer):
    text_error = "Ah ah ah. You didn't say the magic word."

    def validate_text(self, text):
        if text == 'hello':
            raise serializers.ValidationError(ChatTestSerializer.text_error)



from django.conf import settings
from django.core.signals import setting_changed
from django.utils.module_loading import import_string


class ImportDict(dict):
    def __getattribute__(self, item):
        """Whenever accessing an attribute, attempt to import it and override the
        underlying dictionary str value with this import."""
        val = self[item]
        if isinstance(val, str):
            val = import_string(val)
        self[item] = val
        return val


chat_settings = {
    'SERIALIZERS': ImportDict({
        # NOTE: these are referenced like chat_settings.USER (as attribute, not dict key!)
        'USER': 'chit_chat.serializers.ChatUserSerializer',
        'MESSAGE': 'chit_chat.consumer_serializers.ChatMessageSerializer',
    })
}


def load_or_reload_settings(setting=None, **kwargs):
    if setting is not None and setting != 'CKC_CHAT_SERIALIZERS':
        return
    # Override default serializers..
    if hasattr(settings, 'CKC_CHAT_SERIALIZERS'):
        for name, module in settings.CKC_CHAT_SERIALIZERS.items():
            chat_settings['SERIALIZERS'][name] = module


load_or_reload_settings()

setting_changed.connect(load_or_reload_settings)



from collections import OrderedDict

from rest_framework import serializers


#  TODO: This should be stored in the django-ckc package
class PrimaryKeyWriteSerializerReadField(serializers.PrimaryKeyRelatedField):
    def __init__(self, *args, **kwargs):
        read_serializer = kwargs.pop('read_serializer', None)
        assert read_serializer is not None, (
            'PrimaryKeyWriteSerializerReadField must provide `read_serializer` argument.'
        )
        super().__init__(*args, **kwargs)
        self.read_serializer = read_serializer

    # Related fields will not look up any of the object's values for normal primary
    # key serialization. This override forces the lookup of the entire object.
    def use_pk_only_optimization(self):
        return False

    def to_representation(self, value, pk_only=False):
        if pk_only:
            # Returning just the PK so dropdowns can render this as an option in DRF browsable API
            return value.pk
        else:
            # Returning full dict representation for reading normally
            return self.read_serializer(value, context=self.context).data

    def get_choices(self, cutoff=None):
        """Overriding this base method to change to_representation so it passes pk_only=True"""
        queryset = self.get_queryset()
        if queryset is None:
            # Ensure that field.choices returns something sensible
            # even when accessed with a read-only field.
            return {}

        if cutoff is not None:
            queryset = queryset[:cutoff]

        return OrderedDict([
            (
                self.to_representation(item, pk_only=True),
                self.display_value(item)
            )
            for item in queryset
        ])



import random
import string

import factory
from django.contrib.auth import get_user_model
from factory.django import DjangoModelFactory

from chit_chat.models import Room, Message


User = get_user_model()


class UserFactory(DjangoModelFactory):
    class Meta:
        model = User
    email = factory.LazyAttribute(lambda x: f'test_user{"".join(random.sample(string.ascii_letters, 8))}@{"".join(random.sample(string.ascii_letters, 8))}.com')


class RoomFactory(DjangoModelFactory):
    class Meta:
        model = Room

    @factory.post_generation
    def members(self, created, extracted):
        if not created:
            if extracted:
                raise ValueError('Cannot set members on RoomFactory if using .build()')
            return
        self.save()
        for user in extracted if extracted else [UserFactory() for _ in range(random.randint(1, 5))]:
            self.members.add(user)


class MessageFactory(DjangoModelFactory):
    class Meta:
        model = Message
    text = factory.Faker('text')
    user = factory.SubFactory(UserFactory)
    room = factory.SubFactory(RoomFactory)



# Generated by Django 3.2.5 on 2022-03-31 22:34

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('chit_chat', '0003_remove_room_members'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='room',
            name='members_temp',
        ),
        migrations.AddField(
            model_name='room',
            name='members',
            field=models.ManyToManyField(related_name='chat_rooms', through='chit_chat.RoomMembership', to=settings.AUTH_USER_MODEL),
        ),
    ]
